/* HPC University Bi-Weekly Challenge: JavaScript Diffusion, Limited Aggregation
 * Authors: Ishaan Rao and Aaron Weeden, Shodor, 2015
 */

var DEFAULT_NUM_FREE = 200;
var DEFAULT_FREE_COLOR = 'blue';
var DEFAULT_STUCK_COLOR = 'red';
var DEFAULT_NUM_STUCK = 1;
var DEFAULT_PARTICLE_HEIGHT = 5;
var DEFAULT_PARTICLE_WIDTH = 5;
var MAX_INTERVAL_MILLIS = 100;
var BACKGROUND_COLOR = 'white';
var CANVAS_ID = 'canvas';
var STICK_SLIDER_ID = 'stick';
var SPEED_SLIDER_ID = 'speedSlide';
var FREE_PICKER_ID = 'freePick';
var STUCK_PICKER_ID = 'stuckPick';
var FREE_P_ID = 'freeP';
var STUCK_P_ID = 'stuckP';
var STICK_SPAN_ID = 'stickiness';
var PLAY_BUTTON_ID = 'play';
var PAUSE_BUTTON_ID = 'pause';
var STEP_BUTTON_ID = 'step';
var NUM_PARTICLES = DEFAULT_NUM_FREE + DEFAULT_NUM_STUCK;

var NumFree = DEFAULT_NUM_FREE;
var FreeColor = DEFAULT_FREE_COLOR;
var NumStuck = DEFAULT_NUM_STUCK;
var StuckColor = DEFAULT_STUCK_COLOR;
var Particles = new Array(NUM_PARTICLES);
var ParticleHeight = DEFAULT_PARTICLE_HEIGHT;
var ParticleWidth = DEFAULT_PARTICLE_WIDTH;
var StickSlider;
var StickLevel;
var SpeedSlider;
var IntervalMillis;
var MaxIntervalMillis = MAX_INTERVAL_MILLIS;
var Canvas;
var Context;
var Interval;
var Running = false;

onload = function() {
  initialize();

  $('#' + FREE_PICKER_ID).spectrum({
    color: FreeColor,
    change: function(color) {
      if (color.toHexString() != StuckColor) {
        for (var i = 0; i < NUM_PARTICLES; i++) {
          if (Particles[i].color == FreeColor) {
            Particles[i].color = color.toHexString();
          }
        }
        FreeColor = color.toHexString();
        draw();
      }
      else {
        alert('Both colors are the same!! ' +
          'Choose new color for the free particles.');
        $('#' + FREE_PICKER_ID).spectrum('set', FreeColor);
      }
    }
  });
  $('#' + STUCK_PICKER_ID).spectrum({
    color: StuckColor,
    change: function(color) {
      if (color.toHexString() != FreeColor) {
        for (var i = 0; i < NUM_PARTICLES; i++) {
          if (Particles[i].color == StuckColor) {
            Particles[i].color = color.toHexString();
          }
        }
        StuckColor = color.toHexString();
        draw();
      }
      else {
        alert('Both colors are the same!! ' +
          'Choose new color for the stuck particles.');
        $('#' + STUCK_PICKER_ID).spectrum('set', StuckColor);
      }
    }
  });
}

function initialize() {
  Canvas = document.getElementById(CANVAS_ID);
  Context = Canvas.getContext('2d');

  StickSlider = document.getElementById(STICK_SLIDER_ID);
  StickLevel = StickSlider.value;

  SpeedSlider = document.getElementById(SPEED_SLIDER_ID);
  IntervalMillis = MaxIntervalMillis - SpeedSlider.value;

  Particles[0] = {
    x: Canvas.width/2 - ParticleWidth,
    y: Canvas.height/2 - ParticleHeight,
    color: StuckColor
  };

  for (var i = 1; i < NUM_PARTICLES; i++) {
    Particles[i] = {
      x: Math.floor(Math.random() * (Canvas.width/ParticleWidth)) *
        ParticleWidth,
      y: Math.floor(Math.random() * (Canvas.height/ParticleHeight)) *
        ParticleHeight,
      color: FreeColor
    };
  }

  draw();
}

function draw() {
  Context.fillStyle = BACKGROUND_COLOR;
  Context.fillRect(0, 0, Canvas.width, Canvas.height);

  document.getElementById(FREE_P_ID).innerHTML = ': ' + NumFree;
  document.getElementById(STUCK_P_ID).innerHTML = ': ' + NumStuck;
  document.getElementById(STICK_SPAN_ID).innerHTML = 'Stickiness: ' +
    StickLevel + '%';

  for (var i = 0; i < NUM_PARTICLES; i++) {
    Context.fillStyle = Particles[i].color;
    Context.fillRect(Particles[i].x, Particles[i].y, ParticleWidth,
      ParticleHeight);
  }

  getStuck();
  moveRandom();
}

function getStuck() {
  for (var i = 0; i < NUM_PARTICLES; i++) {
    for (var j = 0; j < NUM_PARTICLES; j++) {
      if (Particles[i].x <= Particles[j].x + ParticleWidth &&
          Particles[j].x <= Particles[i].x + ParticleWidth &&
          Particles[i].y <= Particles[j].y + ParticleHeight &&
          Particles[j].y <= Particles[i].y + ParticleHeight &&
          (Particles[i].color == FreeColor) &&
          (Particles[j].color == StuckColor)) {
        Particles[i].stick = Math.floor(Math.random() * 100);
        if (Particles[i].stick < StickLevel) {
          Particles[i].color = StuckColor;
          NumStuck++;
          NumFree--;
        }
      }
    }
  }
}

function moveRandom() {
  for (var i = 0; i < NUM_PARTICLES; i++) {
    if (Particles[i].color == FreeColor) {

      // Get random direction
      var dir = getRandomDir();

      // Check for invalid moves
      for (var j = 0; (dir != '') && (j < NUM_PARTICLES); j++) {
        if ((dir == 'up' &&
              ((Particles[i].y <= 0) ||
                (Particles[j].x == Particles[i].x &&
                  Particles[j].y == Particles[i].y - ParticleHeight &&
                  i != j))) ||
            (dir == 'down' &&
              ((Particles[i].y >= Canvas.height - ParticleHeight) ||
                (Particles[j].x == Particles[i].x &&
                  Particles[j].y == Particles[i].y + ParticleHeight &&
                  i != j))) ||
            (dir == 'right' &&
              ((Particles[i].x >= Canvas.width - ParticleWidth) ||
                (Particles[j].y == Particles[i].y &&
                  Particles[j].x == Particles[i].x + ParticleWidth &&
                  i != j))) ||
            (dir == 'left' &&
              ((Particles[i].x <= 0) ||
                (Particles[j].y == Particles[i].y &&
                  Particles[j].x == Particles[i].x - ParticleWidth &&
                  i != j)))) {
          dir = '';
        }
      }

      // Commit the move
      if (dir == 'up') {
        Particles[i].y -= ParticleHeight;
      }
      else if (dir == 'left') {
        Particles[i].x -= ParticleWidth;
      }
      else if (dir == 'down') {
        Particles[i].y += ParticleHeight;
      }
      else if (dir == 'right') {
        Particles[i].x += ParticleWidth;
      }
    }
  }
}

function getRandomDir() {
    switch(Math.floor(Math.random() * 4)) {
      case 0:
        return('up');
        break;
      case 1:
         return('left');
        break;
      case 2:
        return('down');
        break;
      case 3:
        return('right');
        break;
    }
}

function play() {
  Interval = setInterval('draw()', IntervalMillis);
  Running = true;
  document.getElementById(PLAY_BUTTON_ID).disabled = true;
  document.getElementById(PAUSE_BUTTON_ID).disabled = false;
  document.getElementById(STEP_BUTTON_ID).disabled = true;
}

function pause() {
  clearInterval(Interval);
  Running = false;
  document.getElementById(PLAY_BUTTON_ID).disabled = false;
  document.getElementById(PAUSE_BUTTON_ID).disabled = true;
  document.getElementById(STEP_BUTTON_ID).disabled = false;
}

function timestep() {
  draw();
}

function reset() {
  Running = false;
  clearInterval(Interval);
  NumFree = DEFAULT_NUM_FREE;
  NumStuck = DEFAULT_NUM_STUCK;
  initialize();
  document.getElementById(PAUSE_BUTTON_ID).disabled = true;
  document.getElementById(PLAY_BUTTON_ID).disabled = false;
  document.getElementById(STEP_BUTTON_ID).disabled = false;
}

function changeSpeed() {
  SpeedSlider = document.getElementById(SPEED_SLIDER_ID);
  if (!Running) {
    IntervalMillis = MaxIntervalMillis - (MaxInterval/100 * SpeedSlider.value);
  }
  else if (Running) {
    IntervalMillis = MaxIntervalMillis - SpeedSlider.value;
    clearInterval(Interval);
    Interval = setInterval('draw()', IntervalMillis);
  }
}

function changeStickiness() {
  StickSlider = document.getElementById(STICK_SLIDER_ID);
  StickLevel = StickSlider.value;
  document.getElementById(STICK_SPAN_ID).innerHTML = 'Stickiness: ' +
    StickLevel + '%';
}
